@@ -93,7 +93,7 @@ end)
 let builtin_regexps =
   List.fold_left (fun acc (n, c) -> StringMap.add n (Sedlex.chars c) acc)
     StringMap.empty
-    [
+    ([
      "any", Cset.any;
      "eof", Cset.eof;
      "xml_letter", Cset.letter;
@@ -103,57 +103,9 @@ let builtin_regexps =
      "xml_ideographic", Cset.ideographic;
      "xml_combining_char", Cset.combining_char;
      "xml_blank", Cset.blank;
-     "tr8876_ident_char", Cset.tr8876_ident_char;
-
-     (* Unicode 6.3 categories *)
-     "cc", Unicode63.Categories.cc;
-     "cf", Unicode63.Categories.cf;
-     "cn", Unicode63.Categories.cn;
-     "co", Unicode63.Categories.co;
-     "cs", Unicode63.Categories.cs;
-     "ll", Unicode63.Categories.ll;
-     "lm", Unicode63.Categories.lm;
-     "lo", Unicode63.Categories.lo;
-     "lt", Unicode63.Categories.lt;
-     "lu", Unicode63.Categories.lu;
-     "mc", Unicode63.Categories.mc;
-     "me", Unicode63.Categories.me;
-     "mn", Unicode63.Categories.mn;
-     "nd", Unicode63.Categories.nd;
-     "nl", Unicode63.Categories.nl;
-     "no", Unicode63.Categories.no;
-     "pc", Unicode63.Categories.pc;
-     "pd", Unicode63.Categories.pd;
-     "pe", Unicode63.Categories.pe;
-     "pf", Unicode63.Categories.pf;
-     "pi", Unicode63.Categories.pi;
-     "po", Unicode63.Categories.po;
-     "ps", Unicode63.Categories.ps;
-     "sc", Unicode63.Categories.sc;
-     "sk", Unicode63.Categories.sk;
-     "sm", Unicode63.Categories.sm;
-     "so", Unicode63.Categories.so;
-     "zl", Unicode63.Categories.zl;
-     "zp", Unicode63.Categories.zp;
-     "zs", Unicode63.Categories.zs;
-
-     (* Unicode 6.3 properties *)
-     "alphabetic", Unicode63.Properties.alphabetic;
-     "ascii_hex_digit", Unicode63.Properties.ascii_hex_digit;
-     "hex_digit", Unicode63.Properties.hex_digit;
-     "id_continue", Unicode63.Properties.id_continue;
-     "id_start", Unicode63.Properties.id_start;
-     "lowercase", Unicode63.Properties.lowercase;
-     "math", Unicode63.Properties.math;
-     "other_alphabetic", Unicode63.Properties.other_alphabetic;
-     "other_lowercase", Unicode63.Properties.other_lowercase;
-     "other_math", Unicode63.Properties.other_math;
-     "other_uppercase", Unicode63.Properties.other_uppercase;
-     "uppercase", Unicode63.Properties.uppercase;
-     "white_space", Unicode63.Properties.white_space;
-     "xid_continue", Unicode63.Properties.xid_continue;
-     "xid_start", Unicode63.Properties.xid_start;
-    ]
+     "tr8876_ident_char", Cset.tr8876_ident_char] @
+     Unicode.Categories.list @
+     Unicode.Properties.list)
 
 (* Tables (indexed mapping: codepoint -> next state) *)
 
@@ -189,6 +141,8 @@ let partition_name x =
     Hashtbl.add partitions x s;
     s
 
+(* We duplicate the body for the EOF (-1) case rather than creating
+   an interior utility function. *)
 let partition (name, p) =
   let rec gen_tree = function
     | Lte (i, yes, no) ->
@@ -199,7 +153,10 @@ let partition (name, p) =
         [%expr Char.code (String.get [%e evar (table_name t)] [%e c]) - 1]
   in
   let body = gen_tree (decision_table p) in
-  glb_value name (func [pvar "c", body])
+  glb_value name (func [(pconstr "Some" [pvar "uc"],
+                         [%expr let c = Uchar.to_int uc in [%e body]]);
+                        (pconstr "None" [],
+                         [%expr let c = (-1) in [%e body]])])
 
 (* Code generation for the automata *)
 
@@ -235,13 +192,29 @@ let gen_state lexbuf auto i (trans, final) =
     | Some _ when Array.length trans = 0 -> []
     | Some i -> ret [%expr Sedlexing.mark [%e evar lexbuf] [%e int i]; [%e body ()]]
 
+let gen_recflag auto =
+  (* The generated function is not recursive if the transitions end
+     in states with no further transitions. *)
+  try
+    Array.iter
+      (fun (trans_i, _) ->
+        Array.iter
+          (fun (_, j) ->
+            let (trans_j, _) = auto.(j) in
+            if Array.length trans_j > 0 then raise Exit)
+          trans_i)
+      auto;
+    Nonrecursive
+  with
+    Exit -> Recursive
+
 let gen_definition lexbuf l error =
   let brs = Array.of_list l in
   let auto = Sedlex.compile (Array.map fst brs) in
   let cases = Array.to_list (Array.mapi (fun i (_, e) -> Exp.case (pint i) e) brs) in
   let states = Array.mapi (gen_state lexbuf auto) auto in
   let states = List.flatten (Array.to_list states) in
-  Exp.let_ Recursive states
+  Exp.let_ (gen_recflag auto) states
     (Exp.sequence
        [%expr Sedlexing.start [%e evar lexbuf]]
        (Exp.match_ (appfun (state_fun 0) [evar lexbuf])
